function dx = CSTR(t, x, u, p)
% CSTR A continuously stirred tank reactor dynamical system
%
%   dX = CSTR(T,X,U,P) returns the time-derivatives of each
%       state variable given time (t), current state x(t), input u(t), 
%       and parameter set [p].
% 
%       The system has Nx = 4 state variables and Nu = 4 input variables
%           x = [V,A,B,P],   u = [F_i,F_o,A_i,B_i]
%       The equations depend on a set of Np = 1 parameters
%           p = [k]    
%
    
    % -- Constant parameters
    k = p(1);

    % -- Unpacking state- and control-vars (for readability) --
    V   = x(1);   A = x(2);   B = x(3);   P = x(4);
    F_i = u(1); F_o = u(2); A_i = u(3); B_i = u(4);

    % -- Mass balance equations --
    dx = [
         F_i - F_o;                          % dX_1 / dt
         F_i/V * (A_i - A) -   k * A * B;    % dX_2 / dt
         F_i/V * (B_i - B) - 2*k * A * B;    % dX_3 / dt
        -F_i/V * P         +   k * A * B     % dX_4 / dt
        ];

% --
end