%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (CHEM-E7225) ADVANCED PROCESS CONTROL D
%   #03 - Optimal control with CasADi (Part 2 - Sequential)
%
%   CasADi is a tool to quickly implement and efficiently solve dynamic 
%   optimisation (optimal control) problems. You can also think of it as 
%   a general toolkit for gradient-based optimisation
%
%   CasADI is open-source, LGPL-licensed
%   Download from https://web.casadi.org  (current version: 3.6.4)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all; clear; clc

addpath('~/Documents/MATLAB/casadi')    % --> addpath('~/change_me_with_path_to/casadi-v3.6.4')
import casadi.*

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DYNAMICAL SYSTEM DEFINITIONS

% Control horizon and discretization
tf = 60;                    % Final time
K  = 100;                   % Number of time nodes (control horizon)
dT = tf/K;                  % Sampling time
                            
% Dynamics, initial and reference conditions
Nx = 4;                     % Number of state vars
Nu = 4;                     % Number of controls

% Model parameters
px = [1];

% Initial state
x0 = [5 0.125 0.150 0.150]';

% INTEGRATOR, RK4 (with multishot / simulation horizon)
% In this script, we assume that the control discretisation interval (dT)
%   is different from the simulation discretisation interval (dT_RK4 < dT).
%   We assume that the controls are kept constant between (dT/dT_RK4)
%   simulation time-steps, an assumption known as zero-order hold (ZOH).
%   With this approach, we can have a "better" simulation in the OCP
%   without needing to increase the number of control actions.

x = MX.sym('x',Nx);         % Symbolic state vars
u = MX.sym('u',Nu);         % Symbolic controls
 
K_RK4 = 50;                 % Number and duration of the RK 
dt_RK4 = dT/K_RK4;          % Steps in a sampling interval
             
xplus = x;                                      % Integrator, as a CasADi expression
for k_rk4 = 1:K_RK4                             %   looping through
    xplus = rk4fnc(@CSTR,dt_RK4,xplus,u,px);    %   K_RK4 simulation steps
end                                             % --
F = Function('F', {x,u}, {xplus});              % (Final time) Integrator, as a function


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OPTIMAL CONTROL PROBLEM DEFINITIONS

% == OPTIMIZATION DATA ==

% Construct state-reference functions
xref = x0 + zeros(Nx,K);                            % Sequence of state references:
xref(:,1:K/4)       = 1.00*xref(:,1:K/4);           %             | x0       for k = [   1,     (1/4)K]
xref(:,1+K/4:K/2)   = 1.25*xref(:,1+K/4:K/2);       %  x_ref(k) = { 1.25*x0  for k = [(1/4)K+1, (2/4)K]
xref(:,1+K/2:3*K/4) = 0.75*xref(:,1+K/2:3*K/4);     %             | 0.75*x0  for k = [(2/4)K+1, (3/4)K]
xref(:,1+3*K/4:end) = 1.00*xref(:,1+3*K/4:end);     %             | x0       for k = [(3/4)K+1,    K  ]

% Optimisation parameters
Q  = 1e+0 * eye(Nx);  Q(4,4) = 10;        % State weighting
R  = 1e-4 * eye(Nu);                      % Control weighting
Qf = Q;                                   % Terminal-stage

% - Define lower and upper bounds -
U_lb = zeros(Nu,1);   % (Lower bounds) controls
U_ub = Inf(Nu,1);     % (Upper bounds) controls

% Optimization environment and definition
opti = casadi.Opti();                       % Creates an optimization problem environment

X0 = opti.variable(Nx,1);                    % (Decision vars) states
U  = opti.variable(Nu,K-1);                  % (Decision vars) controls

% - Simulates the system, computes cost function, and store constraints
sumLx = 0;  sumLu = 0;  xk = X0;                    % Stage costs (Lx,Lu) and initial state (xk)
for k = 1:K-1
    sumLx = sumLx + (xk - xref(:,k))' * Q * (xk - xref(:,k));      % Compute state cost, at stage k
    sumLu = sumLu + U(:,k)' * R * U(:,k);                          % Compute action cost, at stage k
    
    xk = F(xk, U(:,k));                             % Computes next state through x_(k+1) = F(.)
end
Lf = (xk - xref(:,K))' * Qf * (xk - xref(:,K));     % Terminal-stage cost

J = (sumLx + sumLu) + Lf;                           % Total cost
% --

% Updates the Opti Stack environment with the objective and constraints
opti.minimize( J )                              % Define the objective function to be minimized
opti.subject_to({ X0 == x0,  ...                % Define the initial state constraint
                  U_lb <= U <= U_ub })          % Define the control bounds

opti.solver('ipopt')                            % Chooses a solver (IPOPT, qpOASES, ...)

opti.set_initial(X0, x0)                        % Sets an initial solution (default: 0)
opti.set_initial(U, rand(Nu,K-1))               % Sets an initial solution (default: 0)

opti.solve();                                   % Executes the solver

% == ==

x0_sol = opti.value(X0);                        % Retrieves the optimal solution of (x0)
u_sol  = opti.value(U);                         % Retrieves the optimal solution of (u)

% Simulates starting from (x0) to obtain the state trajectory
x_sol = nan(Nx,K); x_sol(:,1) = x0_sol;
for k = 1:K-1 
    x_sol(:,k+1) = full( F(x_sol(:,k),u_sol(:,k)) );
end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plotting the simulation results

set(groot,'defaulttextinterpreter','latex');            % These 3 lines sets Latex
set(groot, 'defaultAxesTickLabelInterpreter','latex');  %  as the default interpreter
set(groot, 'defaultLegendInterpreter','latex');         %  for text, axes and legends

fs = 24; ms = 10; lw = 1;
[t_smooth,x_smooth] = Simulate(@CSTR, [0 tf], x0, u_sol, px);

figure(1), clf
for ix = 1:Nx
subplot(Nx,1,ix), hold on
  plot(t_smooth, x_smooth(ix,:), 'k-', 'Linewidth', lw, 'Color', [0.42 0.42 0.42 0.66]')
  stairs((0:K-1)*dT, x_sol(ix,:), 'k.-', 'Linewidth', lw)
  stairs((0:K-1)*dT, xref(ix,:),  'r-', 'Linewidth', lw)
  ylabel("$x_"+num2str(ix)+"$",'FontSize',fs)
  set(gca,'FontSize',fs)
end 
subplot(Nx,1,1 ), title("States")
subplot(Nx,1,Nx)
  xlabel('Time','FontSize',fs)
  legend(["Simulation", "OCP nodes", "Reference"], 'FontSize', 16, 'Box', 'off')

figure(2), clf
for iu = 1:Nu
subplot(Nu,1,iu), hold on
  stairs((0:K-2)*dT, u_sol(iu,:), 'k.-', 'Linewidth', lw)
  ylabel("$u_"+num2str(iu)+"$",'FontSize',fs)
  set(gca,'FontSize',fs)
end 
subplot(Nu,1,1 ), title("Controls")
subplot(Nu,1,Nu), xlabel('Time','FontSize',fs)
